/***********************************************************************

GECCO, a platform for command and control gaming

Copyright (C) 2001 Henrik Brnhielm, Andreas Enblom, Niklas
Hallenfur, Karl Hasselstrm, Henrik Hgerstrm, Oskar Linde, Jing Fu
Zi and Jon slund

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

***********************************************************************/

package gecco.communication.messages;

import java.io.*;
import java.util.*;

import gecco.client.Action;
import gecco.server.core.ActionDefinition;

/**
 *  This message is put in the send queue on the server part of the
 *  communication, when the server calls unitVisible
 *
 * @author     Niklas Hallenfur
 * @author     Karl Hasselstrm
 * @created    June 27, 2001
 */
public class UnitVisibleMessage extends Message {
    /**
     *  Description of the Field
     */
    public int handle;
    /**
     *  Description of the Field
     */
    public String name;
    /**
     *  Description of the Field
     */
    public String type;
    /**
     *  Description of the Field
     */
    public double x;
    /**
     *  Description of the Field
     */
    public double y;
    /**
     *  Description of the Field
     */
    public Set unblockedActions;
    /**
     *  Description of the Field
     */
    public Set blockedActions;
    /**
     *  Description of the Field
     */
    public Map properties;


    /**
     *  Constructor for the UnitVisibleMessage object
     *
     * @param  handle            Description of Parameter
     * @param  name              Description of Parameter
     * @param  type              Description of Parameter
     * @param  x                 Description of Parameter
     * @param  y                 Description of Parameter
     * @param  unblockedActions  Set of unblocked actionDefinitions
     * @param  blockedActions    Set of blocked actionDefinitions
     * @param  properties        Description of Parameter
     */
    public UnitVisibleMessage(int handle, String name, String type,
			      double x, double y, Set unblockedActions,
			      Set blockedActions, Map properties) {
        this.handle = handle;
        this.name = name;
        this.type = type;
        this.x = x;
        this.y = y;
        this.properties = properties;

	// Convert ActionDefinitions to Actions.
        this.unblockedActions = new HashSet();
	if (unblockedActions != null) {
	    Iterator iter = unblockedActions.iterator();
	    while (iter.hasNext()) {
		this.unblockedActions.add(actionDefToAction(iter.next()));
	    }
	}
        this.blockedActions = new HashSet();
	if (blockedActions != null) {
	    Iterator iter = blockedActions.iterator();
	    while (iter.hasNext()) {
		this.blockedActions.add(actionDefToAction(iter.next()));
	    }
	}
    }


    private Action actionDefToAction(Object obj) {

	if (obj instanceof ActionDefinition) {

	    ActionDefinition ad = (ActionDefinition) obj;

	    // Straigtforward parameters.
	    String atype = ad.getName();
	    String adesc = ad.getDescription();
	    boolean instant = ad.isInstantaneous();
	    
	    // Bit more tricky.
	    int paramType;
	    if (ad.getArgumentType().equals("UNIT") ||
		ad.getArgumentType().equals("INSTANT_UNIT")) {
		paramType = Action.PIECE;
	    } else if (ad.getArgumentType().equals("POINT") ||
		       ad.getArgumentType().equals("INSTANT_POINT")) {
		paramType = Action.POSITION;
	    } else if (ad.getArgumentType().equals("VOID") ||
		       ad.getArgumentType().equals("INSTANT_VOID")) {
		paramType = Action.VOID;
	    } else {
		throw new IllegalArgumentException("Broken ActionDefinition!");
	    }
	    
	    // Return the Action.
	    return new Action(atype, adesc, paramType, instant);
	} else {
	    return (Action) obj;
	}
    }


    /**
     *  This method writes the message to the given
     *  <code>DataOutput</code>.
     *
     * @param out The output stream.
     * @exception  IOException  In case something goes wrong.
     */
    public void write(DataOutput out) throws IOException {

        out.writeInt(MessageType.UnitVisible);

        out.writeInt(handle);
        if (name == null) {
            out.writeUTF(new String());
        } else {
            out.writeUTF(name);
        }
        out.writeUTF(type);
        out.writeDouble(x);
        out.writeDouble(y);

        // send unblocked actions
        if (unblockedActions == null) {
            out.writeInt(0);
        } else {
            out.writeInt(unblockedActions.size());
            Iterator it = unblockedActions.iterator();
            while (it.hasNext()) {
                Action a = (Action) it.next();
                out.writeUTF(a.getType());
                out.writeUTF(a.getName());
		out.writeInt(a.getParameterType());
		out.writeBoolean(a.isInstantaneous());
           }
        }

        // send blocked actions
        if (blockedActions == null) {
            out.writeInt(0);
        } else {
            out.writeInt(blockedActions.size());
            Iterator it = blockedActions.iterator();
            while (it.hasNext()) {
                Action a = (Action) it.next();
                out.writeUTF(a.getType());
                out.writeUTF(a.getName());
		out.writeInt(a.getParameterType());
		out.writeBoolean(a.isInstantaneous());
           }
        }

        // send properties
        if (properties == null) {
            out.writeInt(0);
        } else {
            out.writeInt(properties.keySet().size());
            Iterator it = properties.keySet().iterator();
            while (it.hasNext()) {
                String key = (String) it.next();
                out.writeUTF(key);
                out.writeUTF(properties.get(key).toString());
            }
        }
    }


    /**
     *  This factory method reads a message of this type without an
     *  initial ID from the input stream, and returns a
     *  <code>Message</code> object.
     *
     * @param  in  The input stream.
     * @return     An instance of <code>Message</code>.
     */
    public static Message read(DataInput in) throws IOException {

        int handle = in.readInt();
        String name = in.readUTF();
        String type = in.readUTF();
        double x = in.readDouble();
        double y = in.readDouble();

        // get unblocked actions
        int unblockedActionsSize = in.readInt();
        Set unblockedActions = new HashSet();
        for (int i = 0; i < unblockedActionsSize; i++) {
            String atype = in.readUTF();
            String adesc = in.readUTF();
            int paramType = in.readInt();
            boolean instant = in.readBoolean();
            unblockedActions.add(new Action(atype, adesc, paramType, instant));
        }

        // get blocked actions
        int blockedActionsSize = in.readInt();
        Set blockedActions = new HashSet();
        for (int i = 0; i < blockedActionsSize; i++) {
            String atype = in.readUTF();
            String adesc = in.readUTF();
            int paramType = in.readInt();
            boolean instant = in.readBoolean();
            blockedActions.add(new Action(atype, adesc, paramType, instant));
        }

        // get properties
        int propsSize = in.readInt();
        Map props = new TreeMap();
        for (int i = 0; i < propsSize; i++) {
            String pname = in.readUTF();
            String value = in.readUTF();
            props.put(pname, value);
        }

	return new UnitVisibleMessage(handle, name, type, x, y, unblockedActions,
				      blockedActions, props);
    }

}
