/***********************************************************************

GECCO, a platform for command and control gaming

Copyright (C) 2001 Henrik Brnhielm, Andreas Enblom, Niklas
Hallenfur, Karl Hasselstrm, Henrik Hgerstrm, Oskar Linde, Jing Fu
Zi and Jon slund

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

***********************************************************************/

package gecco.server.clock;

import gecco.server.core.*;
import gecco.util.*;

import java.util.*;

/**
 *  This is a virtual game clock. It keeps track of the game time and its
 *  relation to real time. At any moment, the relationship is linear, but seen
 *  over the course of a whole game, the function will be pieced together of
 *  many such linear functions.
 *
 * @author     Karl Hasselstrm
 * @created    July 13, 2001
 */
public class Clock extends SignalThread {
    /**
     *  This is the game time when real time was zero.
     */
    private double zeroTime;
    /**
     *  This constant compensates so that the real time is zero when the clock
     *  was created instead of midnight on January 1, 1970 or whatever the real
     *  system clock uses.
     */
    private long zeroRealTime;
    /**
     *  This is how many times faster the game time runs compared to real time.
     *  This may be zero, in which case the game is paused, mut it may not be
     *  negative.
     */
    private double speed;
    /**
     *  This is the normal, or default speed.
     */
    private double normalSpeed;
    /**
     *  This flag tells whether the game is frozen or not.
     */
    private boolean isFrozen;
    /**
     *  When the game is frozen, this is the current game time.
     */
    private double currentTime;
    /**
     *  This is the set of sleeping threads currently sleeping until a specified
     *  game time.
     */
    private TreeSet sleepers;
    /**
     *  This is the set of sleeping threads that we have interrupted so that
     *  they can adjust their wakeup times.
     */
    private HashSet interrupted;
    /**
     *  This is the minimal speed of the game (to pause it entirely, use <code>
     *  freeze()</code> ).
     */
    private final static double MIN_SPEED = 1e-6;


    /**
     *  Create a game clock.
     *
     * @param  speed        The speed of the game time.
     * @param  normalSpeed  The normal speed of the game time.
     * @param  startTime    The game time should start at this value.
     * @param  isFrozen     <code>true</code> if the game time is initially
     *      frozen, <code>false</code> otherwise.
     */
    public Clock(double startTime, double speed, double normalSpeed, boolean isFrozen) {
        super("Game Clock");

        // Compensate so that real time is zero RIGHT NOW.
        zeroRealTime = System.currentTimeMillis();

        // Make sure the speed is positive.
        if (speed <= 0.0) {
            throw new IllegalArgumentException("Speed of game time must be positive!");
        }

        // Set the linear conversion coefficients.
        this.speed = speed;
        this.normalSpeed = normalSpeed;
        zeroTime = startTime - speed * realTimeNow();

        // Set up time-freeze variables.
        currentTime = startTime;
        this.isFrozen = isFrozen;

        // Create empty sets.
        sleepers = new TreeSet();
        interrupted = new HashSet();
    }


    /**
     *  Sets the speed of the game to its default value.
     */
    public void setNormalSpeed() {
        setSpeed(getNormalSpeed());
    }


    /**
     *  Sets the speed of the game.
     *
     * @param  speed  The new speed of the game time.
     */
    public void setSpeed(double speed) {

        // Make sure the speed is positive.
        speed = Math.max(speed, MIN_SPEED);

        // Note the current game time.
        double currentGameTime = gameTimeNow();

        synchronized (this) {

            // Change the speed.
            this.speed = speed;

            // Recalculate startTime in such a way that the current game
            // time is the same as before despite the speed change.
            zeroTime = currentGameTime - speed * realTimeNow();
        }

        // Nudge the sleep manager thread.
        this.interrupt();
    }


    /**
     *  Gets the Speed attribute of the Clock object
     *
     * @return    The Speed value
     */
    public double getSpeed() {
        synchronized (this) {
            return speed;
        }
    }


    /**
     *  Returns the normal speed of the game time.
     *
     * @return    The normal speed.
     */
    public double getNormalSpeed() {
        synchronized (this) {
            return normalSpeed;
        }
    }


    /**
     *  Gets the Frozen attribute of the Clock object
     *
     * @return    The Frozen value
     */
    public boolean isFrozen() {
        synchronized (this) {
            return isFrozen;
        }
    }


    /**
     *  Converts game time to real time.
     *
     * @param  gameTime  A point in game time.
     * @return           The corresponding point in real time.
     */
    public double realTime(double gameTime) {

        // Make sure the game is not frozen.
        if (isFrozen()) {
            throw new ArithmeticException("Cannot convert game time to " +
                    "real time while game time is frozen!");
        }

        // Convert to real time.
        synchronized (this) {
            return (gameTime - zeroTime) / speed;
        }
    }


    /**
     *  Returns the current value of the real-time clock, in seconds.
     *
     * @return    The real time in seconds.
     */
    public double realTimeNow() {
        return ((double) (System.currentTimeMillis() - zeroRealTime)) / 1000.0;
    }


    /**
     *  Converts real time to game time.
     *
     * @param  realTime  A point in real time.
     * @return           The corresponding point in game time.
     */
    public double gameTime(double realTime) {

        synchronized (this) {

            // If the game is frozen, game time is constant.
            if (isFrozen()) {
                return currentTime;
            }

            // Otherwise, we have to calculate it.
            return zeroTime + speed * realTime;
        }
    }


    /**
     *  Returns the current game time, in seconds.
     *
     * @return    The game time in seconds.
     */
    public double gameTimeNow() {
        return gameTime(realTimeNow());
    }


    /**
     *  Sleep for the specified amount of real time.
     *
     * @param  time                      The number of real seconds to sleep.
     * @exception  InterruptedException  If the thread is interrupted.
     */
    public void realSleep(double time) throws InterruptedException {
        Thread.sleep(Math.round(1000.0 * time));
    }


    /**
     *  Wait on the specified object for (at most) the specified amount of real
     *  time.
     *
     * @param  obj                       The object to wait on.
     * @param  time                      The number of real seconds to wait.
     * @exception  InterruptedException  If the thread is interrupted.
     */
    public void realWait(Object obj, double time) throws InterruptedException {
        obj.wait(Math.max(1, Math.round(1000.0 * time)));
    }


    /**
     *  Sleep for the specified amount of game time.
     *
     * @param  time                      The number of game seconds to sleep.
     * @exception  InterruptedException  If the thread is interrupted.
     */
    public void gameSleep(double time) throws InterruptedException {

        // Sleeping is exactly the same thing as waiting, except we're
        // waiting for someone who'll never come.
        Object godot = new Object();
        synchronized (godot) {
            gameWait(godot, time);
        }
    }


    /**
     *  Wait on an object for (at most) the specified amount of game time.
     *
     * @param  obj                       The object.
     * @param  time                      The number of game seconds to wait.
     * @exception  InterruptedException  If the thread is interrupted.
     */
    public void gameWait(Object obj, double time) throws InterruptedException {
	ReferenceHolder.debug.println("" + Thread.currentThread() + " waits on " + obj +
				      " for at most " + time + " seconds.");

        // Calculate the wakeup time.
	ReferenceHolder.debug.println("Calculating wakeup time.");
        double wakeupTime = gameTimeNow() + time;

        // Add this thread to the list of sleepers.
	ReferenceHolder.debug.println("Adding thread to list.");
        SleepingThread st = new SleepingThread(Thread.currentThread(), wakeupTime);
        synchronized (sleepers) {
            sleepers.add(st);

	    // Nudge the sleep manager thread.
	    this.interrupt();
        }

        // Wait for obj, or until someone interrupts us.
	ReferenceHolder.debug.println("Start waiting.");
        try {
            obj.wait();
        } catch (InterruptedException e) {

            // Check if the InterruptedException was caused by us.
            synchronized (interrupted) {
                if (interrupted.contains(st)) {

                    // Clear the interrupted flag.
                    Thread.interrupted();

                    // Note that we've taken care of the interrupt.
                    interrupted.remove(st);
                } else {

                    // Someone else interrupted this thread, so remove
                    // it from the list of sleepers and rethrow the
                    // InterruptedException.
                    synchronized (sleepers) {
                        sleepers.remove(st);
                    }
                    e.fillInStackTrace();
                    throw e;
                }
            }
        }

        // Remove this thread from the list of sleepers.
	ReferenceHolder.debug.println("Removing thread to list.");
        synchronized (sleepers) {
            sleepers.remove(st);
        }

	ReferenceHolder.debug.println("Leaving gameWait().");
    }


    /**
     *  Freezes or unfreezes the game.
     *
     * @param  freeze  true if the game should freeze, false if it should
     *      unfreeze.
     */
    public void freeze(boolean freeze) {

        synchronized (this) {

            // Note the current game time.
            double currentGameTime = gameTimeNow();

            // Set currentTime or startTime depending on whether we'll
            // be frozen or non-frozen.
            if (freeze) {
                currentTime = currentGameTime;
            } else {
                zeroTime = currentGameTime - speed * realTimeNow();
            }

            // Record our frozen status.
            isFrozen = freeze;
        }

        // Nudge the sleep manager thread.
        this.interrupt();
    }


    /**
     *  Wake up sleeping threads at the times they asked for.
     */
    public void run() {
	ReferenceHolder.debug.println("Entering clock loop.");

        // Loop forever.
        while (true) {

            /*
             * Check for signals.
             */
	    ReferenceHolder.debug.println("Checking signals.");
            switch (SignalThread.signalPending()) {
                case SIGKILL:
                case SIGQUIT:
                    // Stop executing.
                    return;
                case SIGPAUSE:
                    SignalThread.pauseHandler();
                    SignalThread.clearSignal();
                    break;
                default:
                    // Do nothing.
                    SignalThread.clearSignal();
                    break;
            }

            /*
             * Find the thread with the earliest wakeup time.
             */
	    ReferenceHolder.debug.println("Finding earliest thread.");
	    synchronized (sleepers) {
		ReferenceHolder.debug.println(sleepers);
	    }
            SleepingThread first = null;
            try {
                synchronized (sleepers) {
                    first = (SleepingThread) sleepers.first();
                }
            } catch (NoSuchElementException e) {
                // No sleeping threads at all.
            }

            /*
             * If it's due for wake-up, wake it up!
             */
	    ReferenceHolder.debug.println("Should I wake it up?");
            if (first != null && first.wakeupTime <= gameTimeNow()) {

                // Wake up the thread.
                synchronized (sleepers) {
                    synchronized (interrupted) {
                        sleepers.remove(first);
                        interrupted.add(first);
                        first.thread.interrupt();
                    }
                }

                // Start over again.
                continue;
            }

            /*
             * Otherwise, sleep until the time comes.
             */
	    ReferenceHolder.debug.println("No, don't wake it. Sleep.");
            long sleepTime;
            if (isFrozen() || first == null) {

                // If time is frozen or nobody is waiting, sleep for a loooong time.
                sleepTime = Long.MAX_VALUE;
            } else {

                // Otherwise, sleep only as long as necessary.
                double realWakeupTime = realTime(first.wakeupTime);
                double realSleepTime = realWakeupTime - realTimeNow();
                sleepTime = Math.round(1000.0 * realSleepTime);
            }
            try {

                // Try to get some sleep.
                Thread.sleep(Math.max(0, sleepTime));
            } catch (InterruptedException e) {
                // We were woken early, maybe because the speed changed.
            }
        }
    }

}
