/***********************************************************************

GECCO, a platform for command and control gaming

Copyright (C) 2001 Henrik Brnhielm, Andreas Enblom, Niklas
Hallenfur, Karl Hasselstrm, Henrik Hgerstrm, Oskar Linde, Jing Fu
Zi and Jon slund

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

***********************************************************************/

package gecco.server.startup;

import java.awt.Color;
import java.io.*;
import java.util.*;

import gecco.server.automaton.*;
import gecco.server.core.*;

/**
 *  This class loads the settings stored in global.conf.
 *
 * @author     Henrik Brnhielm
 * @author     Karl Hasselstrm
 * @author     Henrik Hgerstrm
 * @created    July 11, 2001
 */
public class GlobalConf {

    /**
     *  The automaton map.
     */
    private AutomatonMap automatonMap;
    /**
     *  The TCP port the server should listen to.
     */
    private int connectionPort;
    /**
     *  The start time of the game.
     */
    private double startTime;
    /**
     *  The normal speed of the game.
     */
    private double normalSpeed;


    /**
     *  The constructor parses the file immediately.
     *
     * @param      in                    The configuration is read from this.
     * @exception  IOException           If there is an I/O error.
     * @exception  SyntaxErrorException  If there is a syntax error.
     * @exception  ParseErrorException   If there is a parse error.
     */
    public GlobalConf(Reader in)
             throws IOException, SyntaxErrorException, ParseErrorException {
        parse(in);
    }


    /**
     *  Gets the automaton map.
     *
     * @return    The automaton map.
     */
    public AutomatonMap getAutomatonMap() {
        return automatonMap;
    }


    /**
     *  Gets the connection port.
     *
     * @return    The connection port.
     */
    public int getConnectionPort() {
        return connectionPort;
    }


    /**
     *  Gets the start time.
     *
     * @return    The start time.
     */
    public double getStartTime() {
        return startTime;
    }


    /**
     *  Gets the normal speed.
     *
     * @return    The normal speed.
     */
    public double getNormalSpeed() {
        return normalSpeed;
    }


    /**
     *  Parses the configuration file.
     *
     * @param  in                        The file stream.
     * @exception  IOException           Description of Exception
     * @exception  SyntaxErrorException  Description of Exception
     * @exception  ParseErrorException   Description of Exception
     */
    private void parse(Reader in)
             throws IOException, SyntaxErrorException, ParseErrorException {

        // Read the file into a data structure
        PropertyTree pt = new PropertyTree();
        pt.parseStream(in);

        // Command line defaults
        PropertyTree defaults = pt.getSingleSubtree("defaults");
        connectionPort = defaults.getSingleInt("port");

        // Clock settings
        PropertyTree clock = pt.getSingleSubtree("clock");
        startTime = clock.getSingleDouble("start_time");
        normalSpeed = clock.getSingleDouble("normal_speed");

        // The map
        PropertyTree mapSettings = pt.getSingleSubtree("map");

        // Size
        int mapX = mapSettings.getSingleInt("width");
        int mapY = mapSettings.getSingleInt("height");

        // Automaton class
        String className = mapSettings.getSingleString("class");
	Class automatonClass;
        try {
            Class cl = Class.forName(className, true, ReferenceHolder.getGameClassLoader());
            Object o = cl.newInstance();
            if (!(o instanceof Automaton)) {
                throw new ParseErrorException(className +
                        " is not an Automaton subclass!");
            }
            automatonClass = cl;
        } catch (ClassNotFoundException e) {
            throw new IOException(e.getMessage());
        } catch (IllegalAccessException e) {
            throw new IOException(e.getMessage());
        } catch (InstantiationException e) {
            throw new IOException(e.getMessage());
        }

        // State definitions
        HashMap colorToState = new HashMap();
        Iterator iter = mapSettings.getSubtreeList("state").iterator();
        while (iter.hasNext()) {
            PropertyTree state = (PropertyTree) iter.next();
            Integer number = new Integer(state.getSingleInt("number"));
            PropertyTree color = state.getSingleSubtree("color");
            int red = color.getSingleInt("red");
            int green = color.getSingleInt("green");
            int blue = color.getSingleInt("blue");
            colorToState.put(new Color(red, green, blue), number);
        }

        // Initial state
        String image = mapSettings.getSingleString("initial_state");
        byte[] state = StartServer.getImage(image);

        // Create the map
        automatonMap = new AutomatonMap(automatonClass, mapX, mapY, state, colorToState);
    }

}
