/***********************************************************************

GECCO, a platform for command and control gaming

Copyright (C) 2001 Henrik Brnhielm, Andreas Enblom, Niklas
Hallenfur, Karl Hasselstrm, Henrik Hgerstrm, Oskar Linde, Jing Fu
Zi and Jon slund

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

***********************************************************************/

package gecco.server.startup;

import java.awt.Color;
import java.io.*;
import java.net.*;
import java.util.*;

import gecco.common.*;
import gecco.communication.server.ClientAmbassador;
import gecco.communication.server.RoleDefinition;
import gecco.server.actionprocessor.ActionProcessor;
import gecco.server.actsofgod.*;
import gecco.server.automaton.*;
import gecco.server.clock.*;
import gecco.server.core.*;
import gecco.server.qmanager.QManager;
import gecco.server.unitmanager.UnitManager;
import gecco.server.visibility.*;
import gecco.util.*;

/**
 *  This class is the startup routine for the server, that parses all
 *  configuration files, and sets everything up.
 *
 * @author     Henrik Brnhielm
 * @author     Karl Hasselstrm
 * @author     Henrik Hgerstrm
 * @created    May 3, 2001
 */
public abstract class StartServer {

    /**
     *  Description of the Field
     */
    private static final String globalConfFileName = "global.conf";
    /**
     *  Description of the Field
     */
    private static final String rolesConfFileName = "roles.conf";
    /**
     *  Description of the Field
     */
    private static final String unitTypesConfFileName = "unittypes.conf";
    /**
     *  Description of the Field
     */
    private static final String unitsConfFileName = "units.conf";
    /**
     *  Description of the Field
     */
    private static final String actsOfGodConfFileName = "actsofgod.conf";
    /**
     *  The name (actually, root directory) of the game.
     */
    private static String game;
    /**
     *  The time update thread.
     */
    private static SignalThread timeUpdateThread;
    /**
     *  The Acts of God thread.
     */
    private static SignalThread actsOfGodThread;
    /**
     *  The queue manager thread.
     */
    private static SignalThread qManagerThread;
    /**
     *  The action processor thread.
     */
    private static SignalThread actionProcessorThread;
    /**
     *  The clock thread.
     */
    private static SignalThread clockThread;


    /**
     *  Starts all the server threads and such.
     *
     * @param  props                      Command line properties.
     * @exception  FileNotFoundException  Description of Exception
     * @exception  IOException            Description of Exception
     * @exception  SyntaxErrorException   Description of Exception
     * @exception  ParseErrorException    Description of Exception
     */
    private static void startup(CommandLineProperties props)
	throws FileNotFoundException, IOException, SyntaxErrorException,
	       ParseErrorException {

	// Is this just a plea for help?
	if (props.getBooleanProperty("-?") ||
	    props.getBooleanProperty("-help") ||
	    props.getBooleanProperty("--help")) {
	    
	    System.out.println();

	    System.out.println("Supported command line arguments are");
	    System.out.println();
	    System.out.println("  -gamedir [directory]");
	    System.out.println("    Starts the server with the game whose main " +
			       "directory is [directory].");
	    System.out.println();
	    System.out.println("  -help, --help or -?");
	    System.out.println("    Prints this help.");
	    System.out.println();
	    System.out.println("  -port [port]");
	    System.out.println("    Starts the server on port [port], overriding " +
			       "the default value in global.conf.");
	    System.out.println();
	    System.out.println("  -version");
	    System.out.println("    Displays the version number of GECCO.");

	    System.out.println();
	    System.exit(0);
	}

	// Or is it just a version number query?
	if (props.getBooleanProperty("-version")) {
	    
	    GeccoVersion gv = new GeccoVersion();
	    System.out.println("GECCO version " + gv);
	    System.out.println();
	    System.exit(0);
	}

	// Set up input/output streams.
	ReferenceHolder.in = new BufferedReader(new InputStreamReader(System.in));
	ReferenceHolder.out = new AutoFlushWriter(new OutputStreamWriter(System.out));
	ReferenceHolder.err = new AutoFlushWriter(new OutputStreamWriter(System.err));
        String logfile = props.getProperty("-log");
        if (logfile == null) {
	    ReferenceHolder.log = new PrintWriter(new NullWriter());
	    ReferenceHolder.out.println("\tLog output will be discarded.");
	} else if (logfile.equals("OUT")) {
	    ReferenceHolder.log = ReferenceHolder.out;
	    ReferenceHolder.out.println("\tLog output redirected to STDOUT.");
	} else if (logfile.equals("ERR")) {
	    ReferenceHolder.log = ReferenceHolder.err;
	    ReferenceHolder.out.println("\tLog output redirected to STDERR.");
	} else {
	    ReferenceHolder.log = new AutoFlushWriter(new FileWriter(logfile));
	    ReferenceHolder.out.println("\tLog output redirected to " + logfile + ".");
	}
        String debugfile = props.getProperty("-debug");
        if (debugfile == null) {
	    ReferenceHolder.debug = new PrintWriter(new NullWriter());
	    ReferenceHolder.out.println("\tDebug output will be discarded.");
	} else if (debugfile.equals("OUT")) {
	    ReferenceHolder.debug = ReferenceHolder.out;
	    ReferenceHolder.out.println("\tDebug output redirected to STDOUT.");
	} else if (debugfile.equals("ERR")) {
	    ReferenceHolder.debug = ReferenceHolder.err;
	    ReferenceHolder.out.println("\tDebug output redirected to STDERR.");
	} else {
	    ReferenceHolder.debug = new AutoFlushWriter(new FileWriter(debugfile));
	    ReferenceHolder.out.println("\tDebug output redirected to " + debugfile + ".");
	}

	// What's the name of the game?
        game = props.getStringProperty("-gamedir", null);
        if (game == null) {
            ReferenceHolder.out.print("No game specified with command-line switch -gamedir.\n" +
				      "Please enter the main directory of a game: ");
            game = ReferenceHolder.in.readLine().trim();
        }

	// Load game name and description.
	File file = new File(game, "name.txt");
	BufferedReader in = new BufferedReader(new FileReader(file));
	ReferenceHolder.setGameName(in.readLine().trim());
	in.close();
	file = new File(game, "description.txt");
	in = new BufferedReader(new FileReader(file));
	StringBuffer buf = new StringBuffer();
	for (String line = in.readLine(); line != null; line = in.readLine()) {
	    buf.append(line);
	    buf.append('\n');
	}
	ReferenceHolder.setGameDescription(buf.toString());
	in.close();

	// Make a class loader to read from classes.jar.
	try {
	    File gameClasses = new File(game, "classes/classes.jar");
	    URL [] gameClassesURL = { gameClasses.toURL() };
	    ReferenceHolder.setGameClassLoader(new URLClassLoader(gameClassesURL));
	} catch (MalformedURLException e) {
	    ReferenceHolder.out.println("Could not find game class files. Exiting.\n");
	    System.exit(1);
	} catch (SecurityException e) {
	    ReferenceHolder.out.println("SecurityException prevents loading of game class files.\n");
	    ReferenceHolder.out.println(e);
	    System.exit(1);
	}

	// Global conf. file
        ReferenceHolder.out.print("Parsing " + globalConfFileName + " . . . ");
        GlobalConf globalConf = new GlobalConf(getConfFile(globalConfFileName));
        ReferenceHolder.out.println("\tdone.");

        // Roles conf. file
        ReferenceHolder.out.print("Parsing " + rolesConfFileName + " . . . ");
        RolesConf rolesConf = new RolesConf(getConfFile(rolesConfFileName));
        ReferenceHolder.out.println("\tdone.");

	// Start the clock (in paused mode :-)
	Clock clock = new Clock(globalConf.getStartTime(), globalConf.getNormalSpeed(),
				globalConf.getNormalSpeed(), true);
	TimeString timeString = new TimeString(clock);
	ReferenceHolder.setClock(clock);
	ReferenceHolder.setTimeString(timeString);

	// Set the automaton map.
        ReferenceHolder.setAutomatonMap(globalConf.getAutomatonMap());

	// Create the unit manager.
        ReferenceHolder.setUnitManager(new UnitManager(rolesConf.getRoles(),
						       rolesConf.getGodseyeRoles()));

	// Create the queue manager.
        ReferenceHolder.setQManager(new QManager());

	// Create the Acts of God scheduler.
        ActsOfGod actsOfGod = new ActsOfGod();

        // Unit types conf. file
        ReferenceHolder.out.print("Parsing " + unitTypesConfFileName + " . . . ");
        UnitTypesConf.parse(getConfFile(unitTypesConfFileName),
			    ReferenceHolder.getUnitManager());
        ReferenceHolder.out.println("\tdone.");

        // Units conf. file
        ReferenceHolder.out.print("Parsing " + unitsConfFileName + " . . . ");
        UnitsConf.parse(getConfFile(unitsConfFileName), ReferenceHolder.getUnitManager());
        ReferenceHolder.out.println("\tdone.");

        // Acts of God conf. file
        ReferenceHolder.out.print("Parsing " + actsOfGodConfFileName + " . . . ");
        ActsOfGodConf.parse(getConfFile(actsOfGodConfFileName), actsOfGod);
        ReferenceHolder.out.println("\tdone.");

        // Get port from command line. If not defined there, take it
        // from global.conf.
        int connectionPort = props.getIntProperty("port", globalConf.getConnectionPort());
        ReferenceHolder.out.println("Attaching communication server to port " + connectionPort + ".");
        ReferenceHolder.setComServer(new ServerImpl());
        Vector roleDefs = new Vector();
	Iterator roleIter = rolesConf.getRoles().iterator();
	while (roleIter.hasNext()) {
	    String role = (String) roleIter.next();
	    roleDefs.add(new RoleDefinition(role,
					    ReferenceHolder.getUnitManager().isRoleCommander(role)));
	}
        Client client = new ClientAmbassador(roleDefs, ReferenceHolder.getComServer(),
					     connectionPort);

	// Create the visibility filter and visibility info object.
        ReferenceHolder.out.println("Initializing visibility filter");
        VisibilityManager visible = new VisibilityManager();
	visible.setClient(client);
        ReferenceHolder.setVisibilityFilter(visible);
        ReferenceHolder.setVisibilityInfo(visible);

	// Start the clock.
        ReferenceHolder.out.println("Starting the clock");
        clockThread = ReferenceHolder.getClock();
        clockThread.start();

	// Create and start the action processor.
        ReferenceHolder.out.println("Starting action processor");
        ReferenceHolder.setActionProcessor(new ActionProcessor());
        actionProcessorThread = new SignalThread(ReferenceHolder.getActionProcessor(),
						 "Action Processor");
        actionProcessorThread.start();

	// Start the queue manager.
        ReferenceHolder.out.println("Starting queue manager");
        qManagerThread = new SignalThread(ReferenceHolder.getQManager(), "Queue Manager");
        qManagerThread.start();

	// Start the Acts of God scheduler.
        ReferenceHolder.out.println("Starting Acts of God scheduler");
        actsOfGodThread = actsOfGod;
        actsOfGodThread.start();

	// Begin sending time update messages.
        ReferenceHolder.out.println("Starting time ticker");
        timeUpdateThread = new SignalThread(new TimeUpdate(), "Time Ticker");
        timeUpdateThread.start();

	// The server is up and running!
        ReferenceHolder.out.println("Server is ready and waiting for incoming connections . . .");

	// Start server command prompt
	CommandPrompt prompt = new CommandPrompt();
	Thread promptThread = new Thread(prompt, "Command Prompt");
	promptThread.start();

	// This thread dies here, but the others go on.
    }


    public static void shutdown() {
	
	// Send SIGQUIT to all running threads.
	timeUpdateThread.sendSignal(SignalThread.SIGQUIT);
	actsOfGodThread.sendSignal(SignalThread.SIGQUIT);
	qManagerThread.sendSignal(SignalThread.SIGQUIT);
	actionProcessorThread.sendSignal(SignalThread.SIGQUIT);
	clockThread.sendSignal(SignalThread.SIGQUIT);

	// Wait for the threads to die.
	try {
	    timeUpdateThread.join();
	    ReferenceHolder.out.println("Time ticker stopped");
	    actsOfGodThread.join();
	    ReferenceHolder.out.println("Acts of God scheduler stopped");
	    qManagerThread.join();
	    ReferenceHolder.out.println("Queue manager stopped");
	    actionProcessorThread.join();
	    ReferenceHolder.out.println("Action processor stopped");
	    clockThread.join();
	    ReferenceHolder.out.println("Clock stopped");
	} catch (InterruptedException e) {
	    ReferenceHolder.out.println("Interrupted while shutting down threads! " +
					"Server may not have shut down cleanly.");
	}
    }


    private static Reader getConfFile(String name)
	throws FileNotFoundException {
	
	File file = new File(game, "conf/" + name);
	return new BufferedReader(new FileReader(file));
    }

    static byte [] getImage(String name)
	throws FileNotFoundException, IOException {
	
	File file = new File(game, "images/" + name);
	InputStream inStream = new BufferedInputStream(new FileInputStream(file));
	byte[] byteArr = new byte[inStream.available()];
	inStream.read(byteArr);
	inStream.close();
	return byteArr;
    }


    /**
     *  The main program for the StartServer class
     *
     * @param  args  The command line arguments
     */
    public static void main(String[] args) {

	// Make sure we have java >= 1.3.0.
	VersionNumber requiredVersion = new VersionNumber(1, 3, 0);
	JavaVersion javaVersion = new JavaVersion();
	if (javaVersion.lessThan(requiredVersion)) {
	    System.out.println("Running Java version " + javaVersion + ".");
	    System.out.println("Java version " + requiredVersion + " or later required. Exiting.");
	    System.exit(1);
	}

	// Use the Pure Java AWT (PJA) awt replacement. This replaces
	// all (or most :-) of awt's native calls with pure Java
	// equivalents, so that we can use awt for off-screen drawing
	// and such without having an X server running.
	System.setProperty ("awt.toolkit", "com.eteks.awt.PJAToolkit");

	// Start up the server threads.
        try {
            startup(new CommandLineProperties(args));
        } catch (FileNotFoundException e) {
            abort(e.getMessage(), 1);
        } catch (IOException e) {
            abort(e.getMessage(), 2);
        } catch (ParseErrorException e) {
            abort(e.getMessage(), 3);
        } catch (SyntaxErrorException e) {
            abort(e.getMessage(), 4);
        }
    }


    /**
     *  Description of the Method
     *
     * @param  message     Description of Parameter
     * @param  exitStatus  Description of Parameter
     */
    private static void abort(String message, int exitStatus) {
        ReferenceHolder.err.println();
        ReferenceHolder.err.println(message);
        ReferenceHolder.err.println();
        ReferenceHolder.err.println("Server aborted.");
        ReferenceHolder.err.println();
        System.exit(exitStatus);
    }
}
