/***********************************************************************

GECCO, a platform for command and control gaming

Copyright (C) 2001 Henrik Brnhielm, Andreas Enblom, Niklas
Hallenfur, Karl Hasselstrm, Henrik Hgerstrm, Oskar Linde, Jing Fu
Zi and Jon slund

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

***********************************************************************/

package gecco.server.startup;

import java.io.*;
import java.util.*;

import gecco.server.core.*;
import gecco.server.unitmanager.*;

/**
 *  This class loads the settings stored in units.conf.
 *
 * @author     Henrik Brnhielm
 * @author     Karl Hasselstrm
 * @author     Henrik Hgerstrm
 * @created    July 11, 2001
 */
public abstract class UnitsConf {

    /**
     *  Parses the file and updates the unit manager.
     *
     * @param      in                    The configuration is read from this.
     * @param      um                    The unit manager.
     * @exception  IOException           If there is an I/O error.
     * @exception  SyntaxErrorException  If there is a syntax error.
     * @exception  ParseErrorException   If there is a parse error.
     */
    public static void parse(Reader in, UnitManager um)
             throws IOException, SyntaxErrorException, ParseErrorException {

        // Read the file into a data structure
        PropertyTree pt = new PropertyTree();
        pt.parseStream(in);

        // Extract information from the data structure
        Iterator iter = pt.getSubtreeList("unit").iterator();
        while (iter.hasNext()) {
            // Get the next unit
            PropertyTree unit = (PropertyTree) iter.next();

            // Create the unit
            String type = unit.getSingleString("type");
            Unit currentUnit = um.createUnit(type);
            currentUnit.setUnitType(type);

            // Read the name
            currentUnit.setName(unit.getSingleString("name"));

            // An optional commander
            List command = unit.getStringList("command");
            if (command.size() > 1) {
                throw new ParseErrorException("More than one commander specified!");
            } else if (command.size() == 1) {
                currentUnit.setCommander((String) command.get(0));
            }

            // Zero or more observers
            List observe = unit.getStringList("observe");
            Iterator iter2 = observe.iterator();
            while (iter2.hasNext()) {
                currentUnit.addObserver((String) iter2.next());
            }

            // Parse the user-defined properties
            iter2 = unit.getSubtreeList("property").iterator();
            while (iter2.hasNext()) {
                PropertyTree property = (PropertyTree) iter2.next();
                String name = property.getSingleString("name");
                String propertyType = property.getSingleString("type");

                // Make sure the property is not already defined as another type
                String oldType = currentUnit.getPropertyType(name);
                if (oldType != null && !oldType.equals(propertyType)) {
                    throw new ParseErrorException("Property " + name + " cannot be redefined to type " + propertyType);
                }

                if (propertyType.equals("STRING")) {
                    currentUnit.setProperty(name, property.getSingleString("value"));
                } else if (propertyType.equals("INT")) {
                    currentUnit.setProperty(name, property.getSingleInt("value"));
                } else if (propertyType.equals("DOUBLE")) {
                    currentUnit.setProperty(name, property.getSingleDouble("value"));
                } else {
                    throw new ParseErrorException("Type must be STRING, INT or DOUBLE.");
                }
            }

            // Read the initial position
            double x = unit.getSingleDouble("init_x");
            double y = unit.getSingleDouble("init_y");
            currentUnit.setPosition(x, y);
            um.initializeUnitPosition(currentUnit);
        }
    }

}
